function make_callback(env, P)
  cbfn_equation(u, t) = callback_equation(u, t, env, P)
  cb_equation         = FunctionCallback(cbfn_equation,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  cbfn_tci(u, t)      = callback_tci(u, t, env, P)
  cb_tci              = FunctionCallback(cbfn_tci,
                                         CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, env, P)
  cb_hrsc             = FunctionCallback(cbfn_hrsc,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  callbackset = CallbackSet(cb_equation, cb_tci, cb_hrsc)
  if !isempty(P.prmsdb["GRHD"]["variables0d_analyze_error"])
    cbfn_analyzed0d_err(u, t) = callback_analyze0d_error(env, P, env.mesh)
    cb                        = FunctionCallback(cbfn_analyzed0d_err,
                                                 CallbackTiming(every_iteration=1,every_dt=0))
    push!(callbackset, cb)
  end
  if !isempty(P.prmsdb["GRHD"]["variables1d_analyze_error"])
    cbfn_analyzed1d_err(u, t) = callback_analyze1d_error(env, P, env.mesh)
    cb                        = FunctionCallback(cbfn_analyzed1d_err,
                                                 CallbackTiming(every_iteration=1,every_dt=0))
    push!(callbackset, cb)
  end
  if !isempty(P.prmsdb["GRHD"]["variables0d_analyze"])
    cbfn_analyzed0d(u, t)     = callback_analyze0d(env, P, env.mesh)
    cb                        = FunctionCallback(cbfn_analyzed0d,
                                                 CallbackTiming(every_iteration=1,every_dt=0))
    push!(callbackset, cb)
  end
  return callbackset
end


#######################################################################
#                              Equation                               #
#######################################################################


function callback_equation(state_u, state_t, env, P)
  # TODO Maybe this should be its own callback
  wrap_dynamic_variables!(env.cache, state_u)
  @unpack t, tm1 = get_global_variables(env.mesh)
  tm1 .= t
  compute_maxspeed(P, P.equation, env.mesh)
  compute_cellmax_v(P, env.mesh)
  compute_atmosphere_mask(env, P, env.mesh)
  compute_entropy_variables(P, P.equation, env.mesh)
end

compute_cellmax_v(P, mesh) = nothing
function compute_cellmax_v(P, mesh::Mesh1d{SpectralElement})
  @unpack max_v = get_static_variables(mesh)
  @unpack cellmax_v = get_cell_variables(mesh)
  for (k,vi) in enumerate(eachcell(mesh, max_v))
    cellmax_v[k] = maximum(vi)
  end
end

compute_atmosphere_mask(env, P, mesh) = nothing
function compute_atmosphere_mask(env, P, mesh::Mesh1d{SpectralElement})
  enforce        = P.prms.c2p_enforce_atm
  enforce_causal = P.prms.c2p_enforce_causal_atm
  (!enforce && !enforce_causal) && return
  if formulation(P) === :valencia1d
    broadcast_volume!(cons2prim_valencia1d, P.equation, env.mesh)
  elseif formulation(P) === :spherical1d
    broadcast_volume!(cons2prim_spherical1d, P.equation, env.mesh)
    broadcast_volume!(determine_atmosphere, P.equation, env.mesh)
    @unpack c2p_reset_atm = get_static_variables(env.mesh)
    @unpack bdry_c2p_reset_atm = get_bdry_variables(env.mesh)
    dg1d.interpolate_face_data!(mesh, c2p_reset_atm, bdry_c2p_reset_atm)
    broadcast_faces!(equalize_atmosphere_spherical1d, P.equation, env.mesh)
  elseif formulation(P) === :rescaled_spherical1d
    broadcast_volume!(cons2prim_rescaled_spherical1d, P.equation, env.mesh)
  elseif formulation(P) === :doublecartoon
    broadcast_volume!(cons2prim_doublecartoon, P.equation, env.mesh)
  elseif formulation(P) === :cartoon
    broadcast_volume!(cons2prim_cartoon, P.equation, env.mesh)
  else
    TODO()
  end
  if P.prms.c2p_dynamic_atm
    broadcast_volume!(determine_atmosphere, P.equation, env.mesh)
  end
  # this controls the effect of the stop_atmosphere_evolution_* methods
  # which trigger when c2p_reset_atm > 0.0 && c2p_freeze_atm > 0.0 at a point
  if enforce
    @unpack c2p_freeze_atm = get_static_variables(env.mesh)
    c2p_freeze_atm .= 1.0 # always on
  elseif enforce_causal
    update_atm_domain_of_dependence!(env.mesh)
  end
end

function update_atm_domain_of_dependence!(mesh::Mesh1d)
  @unpack c2p_freeze_atm, c2p_reset_atm, r = get_static_variables(mesh)
  freeze, reset = c2p_freeze_atm, c2p_reset_atm
  Nx = mesh.element.Npts
  for (cidx,(frz,rst)) in enumerate(zip(eachcell(mesh,freeze),eachcell(mesh,reset)))
    # we need to special case the reset check for when we hit an interface or not
    # ========================================
    # reset check in bulk (requires Nx ≥ 5)
    # --x|x--x---x-----x---x--x|x--
    #        |   |     |
    #       r1  r2    r3
    cell = mesh.tree.cells[cidx]
    # points in bulk
    if Nx ≥ 5
      for i in 3:Nx-2
        r1, r2, r3 = rst[i-1], rst[i], rst[i+1]
        frz[i] = Float64(r1==r2==r3)
      end
    end
    # ========================================
    # reset check close to interface
    # --x|x--x---x-----x---x--x|x--
    #   | |  |   |
    #  r1 r2 r3  r4
    # i = 2
    if dg1d.has_neighbor(cell,Cart1d.Xmin)
      nrst = dg1d.view_neighbor(reset, cell, Cart1d.Xmin, mesh)
      r1, r2, r3, r4 = nrst[Nx], rst[1], rst[2], rst[3]
      frz[2] = Float64(r1==r2==r3==r4)
    else
      r2, r3, r4 = rst[1], rst[2], rst[3]
      frz[2] = Float64(r2==r3==r4)
    end
    # i = Nx-1
    if dg1d.has_neighbor(cell,Cart1d.Xmax)
      nrst = dg1d.view_neighbor(reset, cell, Cart1d.Xmax, mesh)
      r1, r2, r3, r4 = rst[Nx-2], rst[Nx-1], rst[Nx], nrst[1]
      frz[Nx-1] = Float64(r1==r2==r3==r4)
    else
      r1, r2, r3 = rst[Nx-2], rst[Nx-1], rst[Nx]
      frz[Nx-1] = Float64(r1==r2==r3)
    end
    # ========================================
    # reset check on interface
    # --x----x--x|x--x----x--
    #        |  | |  |
    #       r1 r2 r3 r4
    # i = 1
    if dg1d.has_neighbor(cell,Cart1d.Xmin)
      nrst = dg1d.view_neighbor(reset, cell, Cart1d.Xmin, mesh)
      r1, r2, r3, r4 = nrst[Nx-1], nrst[Nx], rst[1], rst[2]
      frz[1] = Float64(r1==r2==r3==r4)
    else
      r3, r4 = rst[1], rst[2]
      frz[1] = Float64(r3==r4)
    end
    # i = Nx
    if dg1d.has_neighbor(cell,Cart1d.Xmax)
      nrst = dg1d.view_neighbor(reset, cell, Cart1d.Xmax, mesh)
      r1, r2, r3, r4 = rst[Nx-1], rst[Nx], nrst[1], nrst[2]
      frz[Nx] = Float64(r1==r2==r3==r4)
    else
      r1, r2 = rst[Nx-1], rst[Nx]
      frz[Nx] = Float64(r1==r2)
    end
  end
  freeze[1] = freeze[end] = 1.0
end
function update_atm_domain_of_dependence!(mesh::Mesh2d)
  @unpack c2p_freeze_atm, c2p_reset_atm, buf_c2p_reset_atm, x, y = get_static_variables(mesh)
  z = y
  freeze, reset, buf_reset = c2p_freeze_atm, c2p_reset_atm, buf_c2p_reset_atm
  # For each point we need to check if all nearest-neighbors have
  # the same value of reset as that point.
  # The fact that our data is duplicated on cell interfaces
  # (and actually quadruplicated on corners in 2d) makes this check a bit involved.
  #
  # To keep the logic below simple we proceed in two steps.
  # 1. Visit each point and fill buf_reset with values from reset, only account for on-site values.
  #   a) If a point lies in the bulk of a cell, the value of reset at that point is copied.
  #   b) If a point lies on an edge and that cell has no neighbor, it can also be copied.
  #   c) If a point lies on an edge and that cell has a neighbor, check both values.
  #      If they agree, then store that value, otherwise store -1.0, which indicates an unknown value.
  #   d) If a point is on a corner, check all neighboring cells for their values.
  #      If they all agree, store it, otherwise store -1.0.
  #
  # 2. Visit each point and fill freeze with nearest-neighbor checks only based on buf_reset.
  #    We built buf_reset so that we only have to visit one nearest-neighbor per direction
  #    in this step. In particular, the value of buf_reset is the same for duplicated points.
  #    a) If a point is in the bulk of a cell, use a normal nearest-neighbor check.
  #    b) If a point is on an edge but there is no neighbor, then just ignore the missing values.
  #    c) If a point is on an edge and there is a neighbor, also include the value from the neighbor.
  #    d) If a point is on a corner, then use all available values for the test.
  #
  # Sketches to illustrate some cases:
  #
  # Legend
  # x ... points
  # C ... current point
  # N ... nearest neighbors used for check
  # c ... point that shares the same coordinates and value as C
  # n ... also nearest neighbor, also shares same coordinates and value as N
  #
  # 2a)
  #   ||  |   |
  # --xx--n---x---
  # --xx--N---x---
  #   ||  |   |
  # --nN--C---N---
  #   ||  |   |
  #   ||  |   |
  # --xx--N---x---
  #   ||  |   |
  #
  # 2c)
  #   |  ||  |   |
  # --x--xx--N---x--
  #   |  ||  |   |
  # --x--nn--c---n--
  # --x--nN--C---N--
  #   |  ||  |   |
  # --x--xx--N---x--
  #   |  ||  |   |
  #
  # 2d)
  #   |  ||  |   |
  # --x--nN--x---x--
  #   |  ||  |   |
  # --n--cc--n---x--
  # --N--cC--N---x--
  #   |  ||  |   |
  # --x--nN--x---x--
  #   |  ||  |   |
  #
  Nx, Nz = mesh.element.Npts, mesh.element.Npts
  Kx, Kz = mesh.tree.dims
  for (cidx,(rst,buf)) in enumerate(zip(eachcell(mesh,reset),eachcell(mesh,buf_reset)))
    cell = mesh.tree.cells[cidx]
    # copy bulk values over
    for i in 2:Nx-1, j in 2:Nz-1
      buf[i,j] = rst[i,j]
    end
    # combine values on edges
    for (xidxs,zidxs,dir,ngbr_xidxs,ngbr_zidxs) in ( (1,      2:Nz-1, Cart2d.Xmin, Nx,     2:Nz-1),
                                                     (Nx,     2:Nz-1, Cart2d.Xmax, 1,      2:Nz-1),
                                                     (2:Nx-1, 1,      Cart2d.Ymin, 2:Nx-1, Nz),
                                                     (2:Nx-1, Nz,     Cart2d.Ymax, 2:Nx-1, 1) )
      if dg1d.has_neighbor(cell, dir)
        nrst = dg1d.view_neighbor(reset, cell, dir, mesh)
        for (i,ii) in zip(xidxs,ngbr_xidxs), (j,jj) in zip(zidxs,ngbr_zidxs)
          buf[i,j] = rst[i,j] != nrst[ii,jj] ? -1.0 : rst[i,j]
        end
      else
        for i in xidxs, j in zidxs
          buf[i,j] = rst[i,j]
        end
      end
    end
    # combine values on corners
    for (i,j,dir1,i1,j1,dir2,i2,j2,dir3,i3,j3) in ( ( 1,  1, Cart2d.Xmin, Nx,  1, Cart2d.Ymin,  1, Nz, Cart2d.Xmin, Nx, Nz),
                                                    (Nx,  1, Cart2d.Xmax, 1,   1, Cart2d.Ymin, Nx, Nz, Cart2d.Xmax,  1, Nz),
                                                    ( 1, Nz, Cart2d.Xmin, Nx, Nz, Cart2d.Ymax,  1,  1, Cart2d.Xmin, Nx,  1),
                                                    (Nx, Nz, Cart2d.Xmax, 1,  Nz, Cart2d.Ymax, Nx,  1, Cart2d.Xmax,  1,  1) )
      r = rst[i,j]
      if dg1d.has_neighbor(cell,dir1)
        nrst = dg1d.view_neighbor(reset, cell, dir1, mesh)
        r != nrst[i1,j1] && (buf[i,j] = -1.0; continue)
      end
      if dg1d.has_neighbor(cell,dir2)
        nrst = dg1d.view_neighbor(reset, cell, dir2, mesh)
        r != nrst[i2,j2] && (buf[i,j] = -1.0; continue)
        # third neighbor is given by cell->dir2->dir3
        ngbr = mesh.tree.cells[cell.neighbors[dir2]]
        if dg1d.has_neighbor(ngbr,dir3)
          nrst = dg1d.view_neighbor(reset, ngbr, dir3, mesh)
          r != nrst[i3,j3] && (buf[i,j] = -1.0; continue)
        end
      end
      buf[i,j] = r
    end
  end

  for (cidx,(frz,buf)) in enumerate(zip(eachcell(mesh,freeze),eachcell(mesh,buf_reset)))
    cell = mesh.tree.cells[cidx]
    # points in bulk
    for i in 2:Nx-1, j in 2:Nz-1
      r1, r2, r3, r4, r5 = buf[i,j], buf[i-1,j], buf[i+1,j], buf[i,j-1], buf[i,j+1]
      frz[i,j] = Float64(r1==r2==r3==r4==r5!=-1.0)
    end
    # points on edges
    for (xidxs,zidxs,dir,ngbr_xidxs,ngbr_zidxs) in ( (1,      2:Nz-1, Cart2d.Xmin, Nx-1,   2:Nz-1),
                                                     (Nx,     2:Nz-1, Cart2d.Xmax, 2,      2:Nz-1),
                                                     (2:Nx-1, 1,      Cart2d.Ymin, 2:Nx-1, Nz-1),
                                                     (2:Nx-1, Nz,     Cart2d.Ymax, 2:Nx-1, 2) )
      if dg1d.has_neighbor(cell, dir)
        nbuf = dg1d.view_neighbor(buf_reset, cell, dir, mesh)
        for (i,ii) in zip(xidxs,ngbr_xidxs), (j,jj) in zip(zidxs,ngbr_zidxs)
          r1 = buf[i,j]
          r2 = dir == Cart2d.Xmin ? nbuf[ii,jj] : buf[i-1,j]
          r3 = dir == Cart2d.Xmax ? nbuf[ii,jj] : buf[i+1,j]
          r4 = dir == Cart2d.Ymin ? nbuf[ii,jj] : buf[i,j-1]
          r5 = dir == Cart2d.Ymax ? nbuf[ii,jj] : buf[i,j+1]
          frz[i,j] = Float64(r1==r2==r3==r4==r5!=-1.0)
        end
      else
        for i in xidxs, j in zidxs
          r1 = buf[i,j]
          r2 = dir == Cart2d.Xmin ? r1 : buf[i-1,j]
          r3 = dir == Cart2d.Xmax ? r1 : buf[i+1,j]
          r4 = dir == Cart2d.Ymin ? r1 : buf[i,j-1]
          r5 = dir == Cart2d.Ymax ? r1 : buf[i,j+1]
          frz[i,j] = Float64(r1==r2==r3==r4==r5!=-1.0)
        end
      end
    end
    # points on corners; no dir3 needed here, because that neighbor only influenced corner point
    # who's info is already contained in buf[i1,j1]
    for (i1,j1,i2,j2,i3,j3,dir1,ii1,jj1,dir2,ii2,jj2) in ( ( 1,  1,    2, 1,   1,    2, Cart2d.Xmin, Nx-1,  1, Cart2d.Ymin,  1, Nz-1),
                                                           (Nx,  1, Nx-1, 1,  Nx,    2, Cart2d.Xmax, 2,     1, Cart2d.Ymin, Nx, Nz-1),
                                                           ( 1, Nz,    2, Nz,  1, Nz-1, Cart2d.Xmin, Nx-1, Nz, Cart2d.Ymax,  1,  2  ),
                                                           (Nx, Nz, Nx-1, Nz, Nx, Nz-1, Cart2d.Xmax, 2,    Nz, Cart2d.Ymax, Nx,  2  ) )
      r1, r2, r3 = buf[i1,j1], buf[i2,j2], buf[i3,j3]
      if dg1d.has_neighbor(cell,dir1)
        nbuf = dg1d.view_neighbor(buf_reset, cell, dir1, mesh)
        r4 = nbuf[ii1,jj1]
      else
        r4 = r1
      end
      if dg1d.has_neighbor(cell,dir2)
        nbuf = dg1d.view_neighbor(buf_reset, cell, dir2, mesh)
        r5 = nbuf[ii2,jj2]
      else
        r5 = r1
      end
      frz[i1,j1] = Float64(r1==r2==r3==r4==r5!=-1.0)
    end
  end
end

# TODO Not implemented yet!
compute_entropy_variables(::Project{:doublecartoon}, equation, mesh) = nothing
compute_entropy_variables(::Project{:doublecartoon}, equation, mesh::Mesh1d{FVElement}) = nothing
compute_entropy_variables(P, equation, mesh) = nothing

function compute_entropy_variables(::Project, equation, mesh)
  @unpack E, Em1, flr_E, flr_Em1,
          EP, invdetJ = get_static_variables(mesh)
  @unpack t, tm1      = get_global_variables(mesh)
  dt       = t[1] - tm1[1]
  Em1     .= E
  flr_Em1 .= flr_E
  dt <= 0.0 && return
  broadcast_volume!(entropy_variables, equation, mesh)
  LL      = layout(mesh)
  EP      = dg1d.vreshape(EP, LL)
  E       = dg1d.vreshape(E, LL)
  flr_E   = dg1d.vreshape(flr_E, LL)
  Em1     = dg1d.vreshape(Em1, LL)
  flr_Em1 = dg1d.vreshape(flr_Em1, LL)
  invdetJ = dg1d.vreshape(invdetJ, LL)
  # compute entropy residual
  @turbo @. flr_Em1 += flr_E
  mul!(EP, mesh.element.D, flr_Em1, 0.5, false)
  @turbo begin
    @. EP .*= invdetJ
    @. EP += (E - Em1) / dt
    @. EP = abs(EP)
  end
end

#######################################################################
#                                HRSC                                 #
#######################################################################


function callback_hrsc(state_u, state_t, env, P)
  compute_hrsc(P, P.equation, env.mesh, P.hrsc)
end

compute_hrsc(P, eq, mesh, hrsc::Nothing) = nothing
compute_hrsc(P, eq, mesh, hrsc::AbstractReconstruction) = nothing

function compute_hrsc(P, equation, mesh, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack max_v, v, ρ, ϵ, v, p, r = get_static_variables(mesh)
  @unpack mu, tmp_mu, cellmax_v   = get_cell_variables(mesh)
  @unpack t, tm1                  = get_global_variables(mesh)
  @unpack av_sensor_abslog_D = P.prms
  @unpack av_u = get_static_variables(mesh)
  if formulation(P) === :doublecartoon
    @unpack rD = get_dynamic_variables(mesh)
    if av_sensor_abslog_D
      @. av_u = log10(abs(rD))
    else
      @. av_u = rD
    end
    fill!(tmp_mu, 0.0)
    HRSC.compute_viscosity!(mu, av_u, cellmax_v, hrsc)
  elseif formulation(P) === :spherical1d
    @unpack D = get_dynamic_variables(mesh)
    if av_sensor_abslog_D
      @. av_u = log10(abs(D))
    else
      @. av_u = D
    end
    fill!(tmp_mu, 0.0)
    HRSC.compute_viscosity!(mu, av_u, cellmax_v, hrsc)
  else
    TODO(formulation(P))
  end
  mu[1] = mu[2] = mu[end] = 0 # ignore boundaries
end

function compute_hrsc(P, equation, mesh, hrsc::HRSC.TCIViscosity)
  @unpack mu, flag = get_cell_variables(mesh)
  HRSC.compute_viscosity!(mu, flag, hrsc)
end

function compute_hrsc(P, equation, mesh, hrsc::HRSC.SmoothedArtificialViscosity)
  compute_hrsc(P, equation, mesh, hrsc.av)
  @unpack mu, mu_m1   = get_cell_variables(mesh)
  @unpack smoothed_mu = get_static_variables(mesh)
  @unpack av_drag     = P.prms
  @. mu_m1 = mu
  @. mu = ((1.0-av_drag)*mu+av_drag*mu_m1)
  hrsc.smoother(smoothed_mu, mu, mesh, false)
end

function compute_hrsc(P, equation, mesh, hrsc::HRSC.EntropyViscosity)
  @unpack E, flr_E, EP, vr = get_static_variables(mesh)
  @unpack mu = get_cell_variables(mesh)
  HRSC.compute_viscosity!(mu, E, flr_E, EP, vr, hrsc)
  mu[1] = mu[2] = mu[end] = 0 # ignore boundaries
end


#######################################################################
#                                 TCI                                 #
#######################################################################


function callback_tci(state_u, state_t, env, P)
  compute_tci(P, P.equation, env.mesh, P.tci)
end

compute_tci(P, eq, mesh, tci::Nothing) = nothing

function compute_tci(P, eq, mesh, tci::TCI.Threshold)
  @unpack c2p_freeze_atm = get_static_variables(mesh)
  @unpack flag = get_cell_variables(mesh)
  TCI.compute_indicator!(flag, c2p_freeze_atm, tci)
end

function compute_tci(P, eq, mesh, tci::TCI.EntropyProduction)
  @warn "Figure out entropy fluxes for special relativistic euler equation!"
  TODO(compute_tci)
  @unpack cache, mesh                = env
  @unpack E, Em1, flx_E, flx_Em1, EP = get_static_variables(cache)
  @unpack flag                       = get_cell_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs     = get_bdry_variables(cache)
  @unpack t, tm1                     = get_global_variables(cache)
  TCI.compute_indicator!(flag, EP, EPjump_lhs, EPjump_rhs,
                         E, Em1, flx_E, flx_Em1, t[1], tm1[1],
                         tci)
end

function compute_tci(P, eq, mesh,
    tci::Union{TCI.Diffusion, TCI.Minmod, TCI.Threshold,
               TCI.ModalDecayAverage, TCI.ModalDecayHighest})
  var = tci_indicator_variable(P, mesh)
  @unpack flag = get_cell_variables(mesh)
  TCI.compute_indicator!(flag, log10.(abs.(var)), tci)
  smooth_flag!(flag, mesh)
end

function tci_indicator_variable(::Project{:doublecartoon}, mesh)
  @unpack rD = get_dynamic_variables(mesh)
  return rD
end
tci_indicator_variable(::Project, mesh) = TODO(formulation(P))

function smooth_flag!(flag, mesh)
  _, K = layout(mesh)
  # 'smooth' flag with neighbouring cells by simple averaging
  tmp_flag = copy(flag)
  periodic = isperiodic(mesh)
  for k = 1:K
    k_lhs, k_rhs = periodic_index(k-1, K), periodic_index(k+1, K)
    if !periodic && k == 1
      tmp_flag[k] = (flag[k] + flag[k_rhs]) / 2
    elseif !periodic && k == K
      tmp_flag[k] = (flag[k_lhs] + flag[k]) / 2
    else
      tmp_flag[k] = (flag[k_lhs] + flag[k] + flag[k_rhs]) / 3
    end
  end
  @. flag = tmp_flag
end


#######################################################################
#                         analyze_error_norm                          #
#######################################################################


callback_analyze0d_error(env, P, mesh) = TODO()
# callback_analyze0d_error(env, P, mesh::Mesh1d{FVElement}) = TODO()
function callback_analyze0d_error(env, P, mesh::Mesh1d)
  vars = P.prmsdb["GRHD"]["variables0d_analyze_error"]
  @unpack ana_err = get_static_variables(mesh)
  for var in vars
    v = get_variable(mesh.cache, Symbol(var))
    v_ref = get_variable(mesh.cache, Symbol(string(var, "_ref")))
    v_err_norm = get_variable(mesh.cache, Symbol(string(var, "_err_norm")))
    @. ana_err = v .- v_ref
    v_err_norm .= dg1d.norm_L2(ana_err, mesh)
  end
end


callback_analyze1d_error(env, P, mesh) = TODO()
# callback_analyze1d_error(env, P, mesh::Mesh1d{FVElement}) = TODO()
function callback_analyze1d_error(env, P, mesh::Mesh1d)
  vars = P.prmsdb["GRHD"]["variables1d_analyze_error"]
  for var in vars
    v = get_variable(mesh.cache, Symbol(var))
    v_ref = get_variable(mesh.cache, Symbol(string(var, "_ref")))
    v_err = get_variable(mesh.cache, Symbol(string(var, "_err")))
    @. v_err = v .- v_ref
  end
end


callback_analyze0d(env, P::Project, mesh) = TODO()
function callback_analyze0d(env, P::Project{:spherical1d}, mesh::Mesh1d)
  vars = P.prmsdb["GRHD"]["variables0d_analyze"]
  if "Mtot" in vars
    @unpack D = get_dynamic_variables(mesh)
    @unpack γrr, γθθ, γϕϕ, r = get_static_variables(mesh)
    @unpack Mtot = get_global_variables(mesh)
    Mtot .= dg1d.integrate(mesh) do i
      D[i] * sqrt(γrr[i]*γθθ[i]*γϕϕ[i]) * 4*π
    end
    if dg1d.is_domain_symmetric(mesh)
      # in this case we integrated the sphere twice, so halving the result
      # should correspond to an averaging of the integral from either side
      Mtot ./= 2
    end
  end
end
