#######################################################################
# doublecartoon formulation
#
# This version is similar to valencia1d, but uses Cartesian coordinates,
# evaluates the equations on the y=z=0 axis and uses the double cartoon
# derivative formulae to convert the y- and z-directional fluxes into
# source terms.
# See dg-notes/tex/cartoon.tex for the derivation of the cartoon formulae.
#######################################################################


@with_signature function maxspeed_doublecartoon(equation::Equation)
  @accepts vx, ρ, ϵ
  @accepts α, γxx
  @unpack eos = equation
  cs2 = eos(SquareSpeedOfSound, Density, ρ, InternalEnergy, ϵ)
  cs = sqrt(abs(cs2))
  γuuxx = 1/γxx
  vxu = γuuxx * vx
  v2 = vxu*vx
  βxu = 0
  λ0 = α*vxu - βxu
  W = 1/sqrt(1-v2)
  λp = α/(1-v2*cs2)*(vxu*(1-cs2)+cs/W*sqrt((1-v2*cs2)*γuuxx-vxu^2*(1-cs2))) - βxu
  λm = α/(1-v2*cs2)*(vxu*(1-cs2)-cs/W*sqrt((1-v2*cs2)*γuuxx-vxu^2*(1-cs2))) - βxu
  max_v = absmax(λm, λ0, λp)
  @returns max_v
end


@with_signature function flux_source_doublecartoon(eq::Equation)
  @accepts D, Sx, τ, ϵ, ρ, p, vx
  @accepts α, βux, βuy, βuz
  @accepts ∂αx, ∂αy, ∂αz
  @accepts ∂βuxx, ∂βuxy, ∂βuxz,
           ∂βuyx, ∂βuyy, ∂βuyz,
           ∂βuzx, ∂βuzy, ∂βuzz
  @accepts γxx, γxy, γxz, γyy, γyz, γzz, sqrtdetγ
  @accepts ∂γxxx, ∂γxxy, ∂γxxz, ∂γxyy, ∂γxyz, ∂γxzz
  @accepts ∂γyxx, ∂γyxy, ∂γyxz, ∂γyyy, ∂γyyz, ∂γyzz
  @accepts ∂γzxx, ∂γzxy, ∂γzxz, ∂γzyy, ∂γzyz, ∂γzzz
  @accepts Kxx, Kxy, Kxz, Kyy, Kyz, Kzz
  @accepts ∂yflx_rDy, ∂zflx_rDz, ∂yflx_rSxy, ∂zflx_rSxz, ∂yflx_rτy, ∂zflx_rτz

  detγ = sqrtdetγ^2
  γuuxx = (γzz*γyy - γyz^2)/detγ
  γuuxy = -(-γyz*γxz + γzz*γxy)/detγ
  γuuxz = (-γyy*γxz + γyz*γxy)/detγ
  γuuyy = (γzz*γxx - γxz^2)/detγ
  γuuyz = -(-γxy*γxz + γyz*γxx)/detγ
  γuuzz = (γyy*γxx - γxy^2)/detγ

  # let
  #   γ = [ γxx  γxy  γxz;
  #         γxy  γyy  γyz;
  #         γxz  γyz  γzz ]
  #   γuu = [ γuuxx  γuuxy  γuuxz;
  #           γuuxy  γuuyy  γuuyz;
  #           γuuxz  γuuyz  γuuzz ]
  #   id = γ * γuu
  #   # display(id)
  #   @show det(γ)*α^2
  # end

  βx = βux*γxx + βuy*γxy + βuz*γxz
  βy = βux*γxy + βuy*γyy + βuz*γyz
  βz = βux*γxz + βuy*γyz + βuz*γzz

  sqrtdetg = α*sqrtdetγ
  gtt = -α^2 + βux*βux*γxx + 2*βux*βuy*γxy + 2*βux*βuz*γxz + βuy*βuy*γyy + 2*βuy*βuz*γyz + βuz*βuz*γzz
  gtx = βx
  gty = βy
  gtz = βz
  gxx = γxx
  gxy = γxy
  gxz = γxz
  gyy = γyy
  gyz = γyz
  gzz = γzz

  detg = -sqrtdetg^2
  guuxx = (gyy*gtt*gzz + 2*gty*gyz*gtz - (gyy*gtz^2 + gyz^2*gtt + gty^2*gzz))/detg
  guuxy = -(gxy*gtt*gzz + gyz*gtx*gtz + gty*gxz*gtz - (gxy*gtz^2 + gyz*gxz*gtt + gty*gtx*gzz))/detg
  guuxz = (gty^2*gxz + gxy*gyz*gtt + gtx*gyy*gtz - (gxy*gty*gtz + gxz*gyy*gtt + gty*gyz*gtx))/detg
  guutx = -(gxy*gyz*gtz + gtx*gyy*gzz + gty*gyz*gxz - (gyz^2*gtx + gxy*gty*gzz + gxz*gyy*gtz))/detg
  guuyy = (gxx*gtt*gzz + 2*gxz*gtx*gtz - (gxx*gtz^2 + gxz^2*gtt + gtx^2*gzz))/detg
  guuyz = -(gxx*gyz*gtt + gxy*gtx*gtz + gty*gxz*gtx - (gyz*gtx^2 + gxx*gty*gtz + gxy*gxz*gtt))/detg
  guuty = (gty*gxz^2 + gxx*gyz*gtz + gxy*gtx*gzz - (gxx*gty*gzz + gxy*gxz*gtz + gyz*gxz*gtx))/detg
  guuzz = (gxx*gyy*gtt + 2*gxy*gty*gtx - (gxx*gty^2 + gxy^2*gtt + gtx^2*gyy))/detg
  guutz = -(gxx*gyy*gtz + gxy*gyz*gtx + gxy*gty*gxz - (gxy^2*gtz + gxx*gty*gyz + gxz*gtx*gyy))/detg
  guutt = (gxx*gyy*gzz + 2*gxy*gyz*gxz - (gxx*gyz^2 + gxy^2*gzz + gxz^2*gyy))/detg

  # let
  #   g = [ gtt gtx  gty  gtz;
  #         gtx gxx  gxy  gxz;
  #         gty gxy  gyy  gyz;
  #         gtz gxz  gyz  gzz ]
  #   guu = [ guutt guutx  guuty  guutz;
  #           guutx guuxx  guuxy  guuxz;
  #           guuty guuxy  guuyy  guuyz;
  #           guutz guuxz  guuyz  guuzz ]
  #   id = g * guu
  #   # display(id)
  #   @show -det(g)
  # end

  vy = vz = 0
  vux = vx*γuuxx + vy*γuuxy + vz*γuuxz
  vuy = vx*γuuxy + vy*γuuyy + vz*γuuyz
  vuz = vx*γuuxz + vy*γuuyz + vz*γuuzz
  v2 = vux*vx + vuy*vy + vuz*vz
  W = 1/sqrt(1-v2)
  uut = W/α
  uux = W*(vux - βux/α)
  uuy = W*(vuy - βuy/α)
  uuz = W*(vuz - βuz/α)

  h = 1 + ϵ + p/ρ
  Tuutt = ρ*h*uut*uut + p*guutt
  Tuutx = ρ*h*uut*uux + p*guutx
  Tuuty = ρ*h*uut*uuy + p*guuty
  Tuutz = ρ*h*uut*uuz + p*guutz
  Tuuxx = ρ*h*uux*uux + p*guuxx
  Tuuxy = ρ*h*uux*uuy + p*guuxy
  Tuuxz = ρ*h*uux*uuz + p*guuxz
  Tuuyy = ρ*h*uuy*uuy + p*guuyy
  Tuuyz = ρ*h*uuy*uuz + p*guuyz
  Tuuzz = ρ*h*uuz*uuz + p*guuzz

  Tudxx = gxx*Tuuxx + gxy*Tuuxy + gxz*Tuuxz + gtx*Tuutx
  Tudyx = gxx*Tuuxy + gxy*Tuuyy + gxz*Tuuyz + gtx*Tuuty
  Tudzx = gxx*Tuuxz + gxy*Tuuyz + gxz*Tuuzz + gtx*Tuutz
  Tudtx = gxx*Tuutx + gxy*Tuuty + gxz*Tuutz + gtx*Tuutt
  Tudxy = gxy*Tuuxx + gyy*Tuuxy + gyz*Tuuxz + gty*Tuutx
  Tudyy = gxy*Tuuxy + gyy*Tuuyy + gyz*Tuuyz + gty*Tuuty
  Tudzy = gxy*Tuuxz + gyy*Tuuyz + gyz*Tuuzz + gty*Tuutz
  Tudty = gxy*Tuutx + gyy*Tuuty + gyz*Tuutz + gty*Tuutt
  Tudxz = gxz*Tuuxx + gyz*Tuuxy + gzz*Tuuxz + gtz*Tuutx
  Tudyz = gxz*Tuuxy + gyz*Tuuyy + gzz*Tuuyz + gtz*Tuuty
  Tudzz = gxz*Tuuxz + gyz*Tuuyz + gzz*Tuuzz + gtz*Tuutz
  Tudtz = gxz*Tuutx + gyz*Tuuty + gzz*Tuutz + gtz*Tuutt
  Tudxt = gtx*Tuuxx + gty*Tuuxy + gtz*Tuuxz + gtt*Tuutx
  Tudyt = gtx*Tuuxy + gty*Tuuyy + gtz*Tuuyz + gtt*Tuuty
  Tudzt = gtx*Tuuxz + gty*Tuuyz + gtz*Tuuzz + gtt*Tuutz
  Tudtt = gtx*Tuutx + gty*Tuuty + gtz*Tuutz + gtt*Tuutt

  src_rD  = 0
  src_rSx = α*(Tudtx*∂βuxx + Tudty*∂βuxy + Tudtz*∂βuxz - α*Tuutt*∂αx + 0.5*((Tuuxx + 2*βux*Tuutx + βux^2*Tuutt)*∂γxxx + (Tuuxy + 2*βux*Tuuty + βuy*βux*Tuutt)*∂γxxy + (Tuuxy + 2*βuy*Tuutx + βuy*βux*Tuutt)*∂γxxy + (Tuuxz + 2*βux*Tuutz + βuz*βux*Tuutt)*∂γxxz + (Tuuxz + 2*βuz*Tuutx + βuz*βux*Tuutt)*∂γxxz + (Tuuyy + 2*βuy*Tuuty + βuy^2*Tuutt)*∂γxyy + (Tuuyz + 2*βuy*Tuutz + βuz*βuy*Tuutt)*∂γxyz + (Tuuyz + 2*βuz*Tuuty + βuz*βuy*Tuutt)*∂γxyz + (Tuuzz + 2*βuz*Tuutz + βuz^2*Tuutt)*∂γxzz))*sqrtdetγ
  src_rτ  = α*(Kxx*Tuuxx + 2*Kxy*Tuuxy + 2*Kxz*Tuuxz + Kyy*Tuuyy + 2*Kyz*Tuuyz + Kzz*Tuuzz + Tuutx*(-∂αx + 2*(Kxx*βux + Kxy*βuy + Kxz*βuz)) + Tuuty*(-∂αy + 2*(Kxy*βux + Kyy*βuy + Kyz*βuz)) + Tuutz*(-∂αz + 2*(Kxz*βux + Kyz*βuy + Kzz*βuz)) + Tuutt*(Kxx*βux^2 + Kyy*βuy^2 + Kzz*βuz^2 + 2*Kxy*βuy*βux + 2*Kxz*βuz*βux + 2*Kyz*βuz*βuy - (βux*∂αx + βuy*∂αy + βuz*∂αz)))*sqrtdetγ

  # # cartoon corrections:
  src_rD  -= ∂yflx_rDy  + ∂zflx_rDz
  src_rSx -= ∂yflx_rSxy + ∂zflx_rSxz
  src_rτ  -= ∂yflx_rτy  + ∂zflx_rτz

  nt = - α
  nx = ny = nz = 0
  Jut = ρ*uut
  Jux = ρ*uux
  Juy = ρ*uuy
  Juz = ρ*uuz

  # computing fluxes
  flx_rDx  = sqrtdetg*Jux
  flx_rSxx = sqrtdetg*Tudxx
  # the - is needed to make it agree with the other commonly used form where
  # - sqrtdetg τ = -sqrtdetg (nt*Tuutt+Jut)
  # - sqrtdetg * ( τ*(vux-βux/α) + p*vux ) = -sqrtdetg (nt*Tuutx+Jux)
  # note: the extra minus in src_rτ has already been taken care of
  flx_rτx  = -sqrtdetg*(nt*Tuutx+Jux)

  @returns flx_rDx, flx_rSxx, flx_rτx, src_rD, src_rSx, src_rτ
end


@with_signature function llf_doublecartoon(equation::Equation)
  @accepts D, flx_rDx, Sx, flx_rSxx, τ, flx_rτx, max_v, p
  @accepts ρ, ϵ, vx, βux, βuy, βuz
  @accepts α, βux, r
  @accepts γxx, γxy, γxz, γyy, γyz, γzz, sqrtdetγ
  @accepts [bdry] bdry_D, bdry_Sx, bdry_τ, bdry_max_v, bdry_vx, bdry_p
  @accepts [bdry] nx

  sqrtdetg = α*sqrtdetγ
  detγ = sqrtdetγ^2
  γuuxx = (γzz*γyy - γyz^2)/detγ
  γuuxy = -(-γyz*γxz + γzz*γxy)/detγ
  γuuxz = (-γyy*γxz + γyz*γxy)/detγ
  γuuyy = (γzz*γxx - γxz^2)/detγ
  γuuyz = -(-γxy*γxz + γyz*γxx)/detγ
  γuuzz = (γyy*γxx - γxy^2)/detγ

  vmax     = absmax(max_v, bdry_max_v)
  bdry_vy  = bdry_vz = 0
  bdry_vux = γuuxx * bdry_vx + γuuxy * bdry_vy + γuuxz * bdry_vz

  rD       = sqrtdetγ * D
  rSx      = sqrtdetγ * Sx
  rτ       = sqrtdetγ * τ
  bdry_rD  = sqrtdetγ * bdry_D
  bdry_rSx = sqrtdetγ * bdry_Sx
  bdry_rτ  = sqrtdetγ * bdry_τ

  bdry_flx_rDx  = sqrtdetg * bdry_D * (bdry_vux - βux/α)
  bdry_flx_rSxx = sqrtdetg * ( bdry_Sx * (bdry_vux - βux/α) + bdry_p )
  bdry_flx_rτx  = sqrtdetg * ( bdry_τ * (bdry_vux - βux/α) + bdry_p * bdry_vux )

  nflx      = nx*flx_rDx
  bdry_nflx = nx*bdry_flx_rDx
  nflx_rD   = LLF(nflx, bdry_nflx, rD, bdry_rD, vmax)

  nflx      = nx*flx_rSxx
  bdry_nflx = nx*bdry_flx_rSxx
  nflx_rSx  = LLF(nflx, bdry_nflx, rSx, bdry_rSx, vmax)

  nflx      = nx*flx_rτx
  bdry_nflx = nx*bdry_flx_rτx
  nflx_rτ   = LLF(nflx, bdry_nflx, rτ, bdry_rτ, vmax)

  @returns [bdry] nflx_rD, nflx_rSx, nflx_rτ
end


@with_signature function bdryllf_doublecartoon_tov(equation::Equation)
  @accepts D, flx_rDx, Sx, flx_rSxx, τ, flx_rτx, max_v, p, r
  @accepts [bdry] nx

  # we assume that we run with a symmetric domain so that we have to impose
  # boundary conditions in the vacuum region
  @assert !(abs(r) ≈ 0)

  # vacuum should show no dynamics, hence, we can just copy the flux
  nflx_rD  = nx*flx_rDx
  nflx_rSx = nx*flx_rSxx
  nflx_rτ  = nx*flx_rτx

  @returns [bdry] nflx_rD, nflx_rSx, nflx_rτ
end


@with_signature function bdryllf_doublecartoon_tov_mirrored(equation::Equation)
  @accepts D, flx_rDx, Sx, flx_rSxx, τ, flx_rτx, max_v, p
  @accepts ρ, ϵ, vx, βux, βuy, βuz
  @accepts α, βux, r
  @accepts γxx, γxy, γxz, γyy, γyz, γzz, sqrtdetγ
  @accepts [bdry] bdry_D, bdry_Sx, bdry_τ
  @accepts [bdry] nx


  if nx > 0 # outer boundary

    # vacuum should show no dynamics, hence, we can just copy the flux
    nflx_rD  = nx*flx_rDx
    nflx_rSx = nx*flx_rSxx
    nflx_rτ  = nx*flx_rτx

  else # nx < 0 # origin

    sqrtdetg = α*sqrtdetγ
    detγ = sqrtdetγ^2
    γuuxx = (γzz*γyy - γyz^2)/detγ
    γuuxy = -(-γyz*γxz + γzz*γxy)/detγ
    γuuxz = (-γyy*γxz + γyz*γxy)/detγ
    γuuyy = (γzz*γxx - γxz^2)/detγ
    γuuyz = -(-γxy*γxz + γyz*γxx)/detγ
    γuuzz = (γyy*γxx - γxy^2)/detγ

    # # apply cons2prim to bdry values
    # We can perform cons2prim analytically in this case, because
    # bdry_Sx = 0 implies bdry_vx = 0. Thus, W = 1.
    bdry_vx = 0
    bdry_vy  = bdry_vz = 0 # also implied by spherical symmetry
    bdry_ρ = bdry_D
    bdry_ϵ = bdry_τ/bdry_ρ
    @unpack eos = equation
    bdry_p = eos(Pressure, Density, bdry_ρ, InternalEnergy, bdry_ϵ)
    bdry_max_v = 0

    vmax     = absmax(max_v, bdry_max_v)
    bdry_vux = γuuxx * bdry_vx + γuuxy * bdry_vy + γuuxz * bdry_vz

    rD       = sqrtdetγ * D
    rSx      = sqrtdetγ * Sx
    rτ       = sqrtdetγ * τ
    bdry_rD  = sqrtdetγ * bdry_D
    bdry_rSx = sqrtdetγ * bdry_Sx
    bdry_rτ  = sqrtdetγ * bdry_τ

    bdry_flx_rDx  = sqrtdetg * bdry_D * (bdry_vux - βux/α)
    bdry_flx_rSxx = sqrtdetg * ( bdry_Sx * (bdry_vux - βux/α) + bdry_p )
    bdry_flx_rτx  = sqrtdetg * ( bdry_τ * (bdry_vux - βux/α) + bdry_p * bdry_vux )

    nflx      = nx*flx_rDx
    bdry_nflx = nx*bdry_flx_rDx
    nflx_rD   = LLF(nflx, bdry_nflx, rD, bdry_rD, vmax)

    nflx      = nx*flx_rSxx
    bdry_nflx = nx*bdry_flx_rSxx
    nflx_rSx  = LLF(nflx, bdry_nflx, rSx, bdry_rSx, vmax)

    nflx      = nx*flx_rτx
    bdry_nflx = nx*bdry_flx_rτx
    nflx_rτ   = LLF(nflx, bdry_nflx, rτ, bdry_rτ, vmax)

  end

  @returns [bdry] nflx_rD, nflx_rSx, nflx_rτ
end


function impose_symmetry_bdry_values_doublecartoon!(mesh::Mesh1d{SpectralElement})

  @unpack cache = mesh
  @unpack D, Sx, τ = get_static_variables(mesh)
  @unpack bdry_D, bdry_Sx, bdry_τ = get_bdry_variables(mesh)

  L = layout(mesh)
  Npts, _ = L
  vD  = dg1d.vreshape(D, L)
  vSx = dg1d.vreshape(Sx, L)
  vτ  = dg1d.vreshape(τ, L)

  bdryidx = 1

  # even variables are interpolated such that derivative vanishes at r=0
  for (bdryvar,var) in ((bdry_D,vD), (bdry_τ,vτ))
    tmp = 0
    for j = 2:Npts
      tmp += mesh.element.D[1,j] * var[j,1]
    end
    bdryvar[bdryidx] = -tmp/mesh.element.D[1,1]
  end
  # odd variables are set to zero at r=0
  bdry_Sx[bdryidx] = 0

end
impose_symmetry_bdry_values_doublecartoon!(mesh) = TODO(mesh)


@with_signature function unscale_conservatives_doublecartoon(equation::Equation)
  @accepts rD, rSx, rτ
  @accepts sqrtdetγ
  D  = rD/sqrtdetγ
  Sx = rSx/sqrtdetγ
  τ  = rτ/sqrtdetγ
  @returns D, Sx, τ
end


@with_signature function doublecartoon_fluxes(eq::Equation)
  @accepts ρ, ϵ, p, vx
  @accepts α, γxx, γxy, γxz, γyy, γyz, γzz, sqrtdetγ

  # This method is tailored to the TOV test case. In particular, it assumes
  # - Sy = Sz = 0
  # - vy = vz = 0
  # - βux = βuy = βuz = 0
  # - γxx = γyy = γzz

  detγ = sqrtdetγ^2
  γuuxx = (γzz*γyy - γyz^2)/detγ
  γuuxy = -(-γyz*γxz + γzz*γxy)/detγ
  γuuxz = (-γyy*γxz + γyz*γxy)/detγ
  γuuyy = (γzz*γxx - γxz^2)/detγ
  γuuyz = -(-γxy*γxz + γyz*γxx)/detγ
  γuuzz = (γyy*γxx - γxy^2)/detγ

  βx = βy = βz = βux = βuy = βuz = 0
  gtt = -α^2 + βux*βux*γxx + 2*βux*βuy*γxy + 2*βux*βuz*γxz + βuy*βuy*γyy + 2*βuy*βuz*γyz + βuz*βuz*γzz
  gtx = βx
  gty = βy
  gtz = βz
  gxx = γxx
  gxy = γxy
  gxz = γxz
  gyy = γyy
  gyz = γyz
  gzz = γzz

  sqrtdetg = α*sqrtdetγ
  detg = -sqrtdetg^2
  guuxx = (gyy*gtt*gzz + 2*gty*gyz*gtz - (gyy*gtz^2 + gyz^2*gtt + gty^2*gzz))/detg
  guuxy = -(gxy*gtt*gzz + gyz*gtx*gtz + gty*gxz*gtz - (gxy*gtz^2 + gyz*gxz*gtt + gty*gtx*gzz))/detg
  guuxz = (gty^2*gxz + gxy*gyz*gtt + gtx*gyy*gtz - (gxy*gty*gtz + gxz*gyy*gtt + gty*gyz*gtx))/detg
  guutx = -(gxy*gyz*gtz + gtx*gyy*gzz + gty*gyz*gxz - (gyz^2*gtx + gxy*gty*gzz + gxz*gyy*gtz))/detg
  guuyy = (gxx*gtt*gzz + 2*gxz*gtx*gtz - (gxx*gtz^2 + gxz^2*gtt + gtx^2*gzz))/detg
  guuyz = -(gxx*gyz*gtt + gxy*gtx*gtz + gty*gxz*gtx - (gyz*gtx^2 + gxx*gty*gtz + gxy*gxz*gtt))/detg
  guuty = (gty*gxz^2 + gxx*gyz*gtz + gxy*gtx*gzz - (gxx*gty*gzz + gxy*gxz*gtz + gyz*gxz*gtx))/detg
  guuzz = (gxx*gyy*gtt + 2*gxy*gty*gtx - (gxx*gty^2 + gxy^2*gtt + gtx^2*gyy))/detg
  guutz = -(gxx*gyy*gtz + gxy*gyz*gtx + gxy*gty*gxz - (gxy^2*gtz + gxx*gty*gyz + gxz*gtx*gyy))/detg
  guutt = (gxx*gyy*gzz + 2*gxy*gyz*gxz - (gxx*gyz^2 + gxy^2*gzz + gxz^2*gyy))/detg

  vy = vz = 0
  vux = vx*γuuxx + vy*γuuxy + vz*γuuxz
  vuy = vx*γuuxy + vy*γuuyy + vz*γuuyz
  vuz = vx*γuuxz + vy*γuuyz + vz*γuuzz
  v2 = vux*vx + vuy*vy + vuz*vz
  W = 1/sqrt(1-v2)
  uut = W/α
  uux = W*vux
  uuy = W*vuy
  uuz = W*vuz
  ux = W*vx
  uy = W*vy
  uz = W*vz

  h = 1 + ϵ + p/ρ
  Tuutt = ρ*h*uut*uut + p*guutt
  Tuutx = ρ*h*uut*uux + p*guutx
  Tuuty = ρ*h*uut*uuy + p*guuty
  Tuutz = ρ*h*uut*uuz + p*guutz
  Tuuxx = ρ*h*uux*uux + p*guuxx
  Tuuxy = ρ*h*uux*uuy + p*guuxy
  Tuuxz = ρ*h*uux*uuz + p*guuxz
  Tuuyy = ρ*h*uuy*uuy + p*guuyy
  Tuuyz = ρ*h*uuy*uuz + p*guuyz
  Tuuzz = ρ*h*uuz*uuz + p*guuzz

  Tudxx = gxx*Tuuxx + gxy*Tuuxy + gxz*Tuuxz + gtx*Tuutx
  Tudyx = gxx*Tuuxy + gxy*Tuuyy + gxz*Tuuyz + gtx*Tuuty
  Tudzx = gxx*Tuuxz + gxy*Tuuyz + gxz*Tuuzz + gtx*Tuutz
  Tudtx = gxx*Tuutx + gxy*Tuuty + gxz*Tuutz + gtx*Tuutt
  Tudxy = gxy*Tuuxx + gyy*Tuuxy + gyz*Tuuxz + gty*Tuutx
  Tudyy = gxy*Tuuxy + gyy*Tuuyy + gyz*Tuuyz + gty*Tuuty
  Tudzy = gxy*Tuuxz + gyy*Tuuyz + gyz*Tuuzz + gty*Tuutz
  Tudty = gxy*Tuutx + gyy*Tuuty + gyz*Tuutz + gty*Tuutt
  Tudxz = gxz*Tuuxx + gyz*Tuuxy + gzz*Tuuxz + gtz*Tuutx
  Tudyz = gxz*Tuuxy + gyz*Tuuyy + gzz*Tuuyz + gtz*Tuuty
  Tudzz = gxz*Tuuxz + gyz*Tuuyz + gzz*Tuuzz + gtz*Tuutz
  Tudtz = gxz*Tuutx + gyz*Tuuty + gzz*Tuutz + gtz*Tuutt
  Tudxt = gtx*Tuuxx + gty*Tuuxy + gtz*Tuuxz + gtt*Tuutx
  Tudyt = gtx*Tuuxy + gty*Tuuyy + gtz*Tuuyz + gtt*Tuuty
  Tudzt = gtx*Tuuxz + gty*Tuuyz + gtz*Tuuzz + gtt*Tuutz
  Tudtt = gtx*Tuutx + gty*Tuuty + gtz*Tuutz + gtt*Tuutt

  nt = - α
  nx = ny = nz = 0

  Jut = ρ*uut
  Jux = ρ*uux
  Juy = ρ*uuy
  Juz = ρ*uuz

  # the formulae here are incomplete; the missing factors of 1/x or ∂x
  # are accounted for in impose_symmetry_doublecartoon! below
  ∂yflx_rDy = sqrtdetg * Jux
  ∂zflx_rDz = sqrtdetg * Jux
  ∂yflx_rSxy = sqrtdetg * (-Tudyy + Tudxx)
  ∂zflx_rSxz = sqrtdetg * (-Tudzz + Tudxx)
  ∂yflx_rτy = - sqrtdetg * (nt * Tuutx + Jux)
  ∂zflx_rτz = - sqrtdetg * (nt * Tuutx + Jux)

  @returns ∂yflx_rDy, ∂zflx_rDz, ∂yflx_rSxy, ∂zflx_rSxz, ∂yflx_rτy, ∂zflx_rτz
end


function impose_symmetry_doublecartoon!(P::Project{:doublecartoon}, mesh::Mesh)

  P.prmsdb["GRHD"]["doublecartoon_disable_derivatives"] && return
  broadcast_volume!(doublecartoon_fluxes, P.equation, mesh)

  @unpack ∂yflx_rDy, ∂zflx_rDz, ∂yflx_rSxy, ∂zflx_rSxz, ∂yflx_rτy, ∂zflx_rτz = get_static_variables(mesh)
  @unpack buf_cartoon, x = get_static_variables(mesh)

  if singularity_on_grid(mesh)
    # when x=0 is on the grid we need to apply l'Hospital's rule to take the limit
    # as a result the factors of 1/x are replaced by ∂x in the cartoon formulae
    # because of that we need to special case here and compute derivatives
    is = findall(xi -> xi ≈ 0.0, x)
    N = length(x)
    for var in (∂yflx_rDy, ∂zflx_rDz, ∂yflx_rSxy, ∂zflx_rSxz, ∂yflx_rτy, ∂zflx_rτz)
      dg1d.differentiate!(buf_cartoon, var, mesh)
      @. var /= x
      for i in is
        var[i] = buf_cartoon[i]
      end
    end
  else
    for var in (∂yflx_rDy, ∂zflx_rDz, ∂yflx_rSxy, ∂zflx_rSxz, ∂yflx_rτy, ∂zflx_rτz)
      @. var /= x
    end
  end

  return
end


@with_signature function fv_bdry_flux_doublecartoon_tov_mirrored(equation::Equation)
  @accepts rD, rSx, rτ, flx_rDx, flx_rSxx, flx_rτx
  @accepts [bdry] nx
  if nx > 0 # outer boundary
    # vacuum should show no dynamics, hence, we can just copy over the values
    bdry_rD  = rD
    bdry_rSx = rSx
    bdry_rτ  = rτ
    nflx_rD  = nx*flx_rDx
    nflx_rSx = nx*flx_rSxx
    nflx_rτ  = nx*flx_rτx
  else
    # reflect at center
    bdry_rD  = rD
    bdry_rSx = -rSx
    bdry_rτ  = rτ
    bdry_flx_rDx  = -flx_rDx
    bdry_flx_rSxx = flx_rSxx
    bdry_flx_rτx  = -flx_rτx
    nflx_rD  = bdry_flx_rDx
    nflx_rSx = bdry_flx_rSxx
    nflx_rτ  = bdry_flx_rτx
  end
  @returns [bdry] bdry_rD, bdry_rSx, bdry_rτ, nflx_rD, nflx_rSx, nflx_rτ
end


#######################################################################
#                                 LDG                                 #
#######################################################################


@with_signature function ldg_nflux_doublecartoon(eq::Equation)
  @accepts        rD, rSx, rτ
  @accepts [bdry] bdry_D, bdry_Sx, bdry_τ, sqrtdetγ
  @accepts [bdry] nx
  bdry_rD, bdry_rSx, bdry_rτ = sqrtdetγ*bdry_D, sqrtdetγ*bdry_Sx, sqrtdetγ*bdry_τ
  nflx_rD  = 0.5*nx*(rD+bdry_rD)
  nflx_rSx = 0.5*nx*(rSx+bdry_rSx)
  nflx_rτ  = 0.5*nx*(rτ+bdry_rτ)
  @returns [bdry] nflx_rD, nflx_rSx, nflx_rτ
end


@with_signature function ldg_bdryllf_doublecartoon(eq::Equation)
  @accepts        rD, rSx, rτ
  @accepts [bdry] nx
  nflx_rD  = nx*rD
  nflx_rSx = nx*rSx
  nflx_rτ  = nx*rτ
  @returns [bdry] nflx_rD, nflx_rSx, nflx_rτ
end


@with_signature function av_flux_covariant_doublecartoon(equations::Equation)
  @accepts rhs_rD, rhs_rSx, rhs_rτ, flx_rDx, flx_rSxx, flx_rτx, ldg_rD, ldg_rSx, ldg_rτ, smoothed_mu
  flx_rDx  += -smoothed_mu * (rhs_rD  - ldg_rD)
  flx_rSxx += -smoothed_mu * (rhs_rSx - ldg_rSx)
  flx_rτx  += -smoothed_mu * (rhs_rτ  - ldg_rτ)
  @returns flx_rDx, flx_rSxx, flx_rτx
end


@with_signature function av_nflux_covariant_doublecartoon(eq::Equation)
  @accepts        rhs_rD, rhs_rSx, rhs_rτ, ldg_rD, ldg_rSx, ldg_rτ, smoothed_mu, r
  @accepts [bdry] bdry_rhs_rD, bdry_rhs_rSx, bdry_rhs_rτ, bdry_ldg_rD, bdry_ldg_rSx, bdry_ldg_rτ, bdry_smoothed_mu
  @accepts [bdry] nflx_rD, nflx_rSx, nflx_rτ, nx

  nflx       = -nx*(rhs_rD - ldg_rD)
  bdry_nflx  = -nx*(bdry_rhs_rD - bdry_ldg_rD)
  nflx_rD    += 0.5 * (smoothed_mu*nflx + bdry_smoothed_mu*bdry_nflx)

  nflx       = -nx*(rhs_rSx - ldg_rSx)
  bdry_nflx  = -nx*(bdry_rhs_rSx - bdry_ldg_rSx)
  nflx_rSx   += 0.5 * (smoothed_mu*nflx + bdry_smoothed_mu*bdry_nflx)

  nflx       = -nx*(rhs_rτ - ldg_rτ)
  bdry_nflx  = -nx*(bdry_rhs_rτ - bdry_ldg_rτ)
  nflx_rτ    += 0.5 * (smoothed_mu*nflx + bdry_smoothed_mu*bdry_nflx)

  @returns [bdry] nflx_rD, nflx_rSx, nflx_rτ
end


@with_signature function stop_atmosphere_evolution_doublecartoon(eq::Equation)
  @accepts rhs_rD, rhs_rSx, rhs_rτ, c2p_reset_atm, c2p_freeze_atm
  if c2p_reset_atm > 0 && c2p_freeze_atm > 0
    rhs_rD = rhs_rSx = rhs_rτ = 0.0
  end
  @returns rhs_rD, rhs_rSx, rhs_rτ
end
