mutable struct Equation{T_EoS<:AbstractEquationOfState} <: AbstractEquation
  eos::T_EoS
  cold_eos::Polytrope
  atm_density::Float64
  atm_factor::Float64
  atm_threshold::Float64
  c2p_threshold_velocity_limiter::Float64
  c2p_set_atmosphere_on_failure::Bool
end


struct Project{Formulation,
               T_Eq    <:AbstractEquation,
               T_HRSC  <:Maybe{HRSC.AbstractHRSC},
               T_TCI   <:Maybe{TCI.AbstractTCI},
               T_Prms}
  equation::T_Eq
  hrsc::T_HRSC
  tci::T_TCI
  prms::T_Prms
  prmsdb::dg1d.ParameterDB
end
Project(formulation, fields...) = Project{formulation,typeof.(fields[1:end-1])...}(fields...)


formulation(::Project{F}) where F = F


struct SubcellDG{T_Mesh<:Mesh{<:SpectralElement}}
  mesh::T_Mesh
  Mcube::Array{Float64,3}
  Scube::Array{Float64,3}
  Bcube::Array{Float64,3}
  M::Matrix{Float64}
  S::Matrix{Float64}
  B::Matrix{Float64}
  nflx::Vector{Float64}
end
function SubcellDG(mesh::Mesh1d{<:SpectralElement})
  @assert mesh.element.quadr ∈ (:LGL, :LGL_lumped)
  M = similar(mesh.element.invM)
  S = similar(mesh.element.invM)
  B = similar(mesh.element.invM)
  Mcube = compute_mass_matrix_cube(mesh)
  Scube = compute_stiffness_matrix_cube(mesh)
  Bcube = compute_boundary_matrix_cube(mesh)
  nflx = zeros(Float64, size(B,2))
  return SubcellDG(mesh, Mcube, Scube, Bcube, M, S, B, nflx)
end


# function bump_function(x, l, r)
#   y = (r-l)*(x+1)/2 + l
#   Φ = (1 - tanh((2*x-1)/(2*x(x-1))))/2
#   return Φ
# end
# function derivative_bump_function(x, l, r)
#   y = (r-l)*(x+1)/2 + l
#   y′ = (r-l)/2
#   x2 = x^2
#   Φ′ = ((1 - 2*x + 2*x2)*sech((1 - 2*x)/(2*x - 2*x2))^2)/(4*(-1 + x)^2*x^2)
#   return Φ′*y′
# end


function compute_mass_matrix_cube(mesh::Mesh1d{<:SpectralElement})
  @unpack Npts, w = mesh.element
  Mcube = zeros(Float64, (Npts,Npts,Npts))
  for n in 1:Npts
    Mcube[n,n,n] = w[n]
  end
  return Mcube
end
function compute_stiffness_matrix_cube(mesh::Mesh1d{<:SpectralElement})
  @unpack Npts, w, z = mesh.element
  Scube = zeros(Float64, (Npts,Npts,Npts))
  for i in 1:Npts, j in 1:Npts, n in 1:Npts
    zn, wi = z[n], w[i]
    ∂lj_n = dg1d.derivative_lagrange_poly(j, zn, z)
    Scube[i,j,n] = wi*∂lj_n * (i==n)
  end
  return Scube
end
function compute_boundary_matrix_cube(mesh::Mesh1d{<:SpectralElement})
  @unpack Npts, w, z = mesh.element
  Bcube = zeros(Float64, (Npts,Npts,Npts))
  subz = zeros(Float64, Npts+1)
  subz[1], subz[end] = z[1], z[end]
  for n in 1:Npts-1
    subz[n+1] = z[n]+(z[n+1]-z[n])/2
  end
  for i in 1:Npts, j in 1:Npts, n in 1:Npts
    subz_m, subz_p = subz[n], subz[n+1]
    li_m, lj_m = dg1d.lagrange_poly(i, subz_m, z), dg1d.lagrange_poly(j, subz_m, z)
    li_p, lj_p = dg1d.lagrange_poly(i, subz_p, z), dg1d.lagrange_poly(j, subz_p, z)
    Bcube[i,j,n] = li_p*lj_p - li_m*lj_m
  end
  return Bcube
end


@inline function compute_mass_matrix!(sdg::SubcellDG, vmask)
  @unpack M, Mcube = sdg
  @unpack Npts = sdg.mesh.element
  for j in 1:Npts, i in j:Npts
    mij = 0.0
    @turbo for k in 1:Npts
      mij += Mcube[i,j,k] * (1.0 - vmask[k])
    end
    M[i,j] = mij
    if i > j
      M[j,i] = mij
    end
  end
end
@inline function compute_stiffness_matrix!(sdg::SubcellDG, vmask)
  @unpack S, Scube = sdg
  @unpack Npts = sdg.mesh.element
  @turbo for j in 1:Npts, i in 1:Npts
    sij = 0.0
    for k in 1:Npts
      sij += Scube[i,j,k] * (1.0 - vmask[k])
    end
    S[i,j] = sij
  end
end
@inline function compute_boundary_matrix!(sdg::SubcellDG, vmask)
  @unpack B, Bcube = sdg
  @unpack Npts = sdg.mesh.element
  @turbo for j in 1:Npts, i in 1:Npts
    bij = 0.0
    for k in 1:Npts
      bij += Bcube[i,j,k] * (1.0 - vmask[k])
    end
    B[i,j] = bij
  end
end


function dg1d.compute_rhs_weak_form!(rhs, f, nf, sdg::SubcellDG, mask)
  @unpack M, S, B, nflx, mesh = sdg
  @unpack w = mesh.element
  @unpack invdetJ = get_static_variables(mesh.cache)
  Npts, K = layout(mesh)
  nf_lhs = view(nf, 1:2:2*K)
  nf_rhs = view(nf, 2:2:2*K)
  for k in 1:K
    cidxs   = cellindices(mesh, k)
    v_mask = view(mask, cidxs)
    s_mask = sum(v_mask)
    v_rhs  = view(rhs, cidxs)
    if s_mask ≈ Npts
      v_rhs .= 0.0
      continue
    elseif s_mask ≈ 0.0
      continue
    end
    # TODO With lumping approx we can skip mass matrix computation and also divide out
    # the quadrature weight in stiffness matrix computation
    # compute_mass_matrix!(sdg, v_mask)
    compute_stiffness_matrix!(sdg, v_mask)
    compute_boundary_matrix!(sdg, v_mask)
    v_f    = view(f, cidxs)
    v_invJ = view(invdetJ, cidxs)
    mul!(v_rhs, transpose(S), v_f)
    nflx[1]   = -nf_lhs[k]
    @views nflx[2:end-1] .= v_f[2:end-1]
    nflx[end] = nf_rhs[k]
    mul!(v_rhs, B, nflx, -1.0, 1.0)
    @turbo for i in 1:Npts
      v_rhs[i] *= (1.0 - v_mask[i]) * v_invJ[i] / w[i]
    end
  end
  return
end
function dg1d.compute_rhs_weak_form!(rhs, f, s, nf, sdg::SubcellDG, mask)
  compute_rhs_weak_form!(rhs, f, nf, sdg, mask)
  Npts, K = layout(sdg.mesh)
  for k in 1:K
    cidxs   = cellindices(sdg.mesh, k)
    v_mask  = view(mask, cidxs)
    s_mask = sum(v_mask)
    s_mask ≈ Npts && continue
    s_mask ≈ 0.0 && continue
    v_rhs   = view(rhs, cidxs)
    v_s     = view(s, cidxs)
    v_rhs .+= v_s .* (1 .- v_mask)
  end
  return
end
