export ScalarEq
module ScalarEq


using dg1d
using LoopVectorization


include("types.jl")
include("equation.jl")
include("rhs.jl")
include("initialdata.jl")
include("callbacks.jl")
include("setup.jl")


#######################################################################
#                 dg1d project interface definitions                  #
#######################################################################


dg1d.project_type(::Val{:ScalarEq}) = Project

# TODO Remove bdryconds
dg1d.rhs!(env::Environment, P::Project, bdryconds) = rhs!(env, env.mesh, P, P.hrsc)

dg1d.timestep!(env::Environment, P::Project) = timestep(env, P, env.mesh, P.hrsc)

dg1d.@parameters ScalarEq begin

  """
  Available options:
  - `advection`
  - `burgers`
  - `balanced_burgers`
  """
  equation = "advection"
  @check equation in [ "advection", "burgers", "balanced_burgers" ]

  """
  Advection speed:
  - 1d: [ vx ]
  - 2d: [ vx, vy ]
  """
  advection_v = [ 1.0 ]
  @check length(advection_v) in [ 1, 2 ]

  """
  Initial data configuration. Available options:
  - `sine`
  - `sine_squared`
  - `sine_squared_vs_constant`
  - `periodic_kink`
  - `bump`
  - `gaussian`
  - `smoothed_bump`
  """
  id = "sine"
  @check id in [ "sine", "sine_squared_vs_constant", "periodic_kink",
                 "bump", "gaussian", "smoothed_bump" ]

  "Period of sine initial data"
  id_sine_period = 1

  "Phase of sine initial data"
  id_sine_phase = 0.0

  "Period of sine squared initial data"
  id_sine_squared_period = 1

  "Phase of sine squared initial data"
  id_sine_squared_phase = 0.0

  """Period of the sine squared that fills half of the domain. The other half is filled by
  a constant and attached smoothly"""
  id_sine_squared_vs_constant_period = 1

  """Period of the sine squared that fills one quater of the domain.
  The number of periods of the periodic kink that fills half of the domain.
  The other half is filled by a constant that is attached non-smoothly but continuous"""
  id_periodic_kink_period = 1

  "RMS width of gaussian"
  id_gaussian_width = 1.0

  "Amplitude of gaussian"
  id_gaussian_amplitude = 1.0

  "Amplitude of bump"
  id_bump_height = 1.0

  "Width of bump"
  id_bump_width = 1.0

  """
  Boundary conditions. Available options:
  - `periodic`
  - `no_inflow`
  - `from_id`
  """
  bc = "periodic"
  @check bc in [ "periodic", "no_inflow", "from_id" ]

  """
  Approximation scheme for 2nd derivative of artificial viscosity regularization.
  Available options:
  - ldg ... local discontinuous Galerkin
  - fd ... finite difference
  """
  av_derivative_scheme = "ldg"
  @check av_derivative_scheme in [ "ldg", "fd" ]

  """
  Determines how quickly artificial viscosity can alter between time steps.
  If `av_drag = 0.0` then no dragging occurs.
  If `av_drag = 1.0` then viscosity is frozen to its initial value.
  """
  av_drag = 0.0
  @check 0.0 <= av_drag <= 1.0

  """
  The HRSC method used for FV discretization.
  Note: For DG methods use the `HRSC` parameter section.
  - `fv_central` ... Riemann solver free FV method, cf. Hesthaven book 2018, sec. 6.3
  - `slope_limiter` ... slope limited HRSC, cf. Hesthaven book 2018, sec. 10.2
  """
  hrsc = "fv_central"
  @check hrsc in [ "fv_central", "slope_limiter" ]

  """
  Available options:
  - `none`
  - `minmod` [1]
  - `muscl` [1] ... monotone upstream-centered limiter
  - `superbee` [1]
  - `van_albada` [1]
  - `van_leer` [1]
  - `tvb` [1] ... total variation bounded

  [1] Hesthaven 2018 book, p. 268, Script 10.10
  """
  slope_limiter_method = "muscl"
  @check slope_limiter_method in [ "none", "minmod", "muscl", "superbee", "van_leer", "tvb" ]

  """
  Slope limiting parameter for MUSCL scheme, cf. Hesthaven book 2008, sec 10.2
  """
  slope_limiter_tvb_M = 150.0
  @check slope_limiter_tvb_M > 0.0

  """
  Slope limiting parameter for MUSCL scheme, cf. Hesthaven book 2008, sec 10.2
  """
  muscl_omega = 0.0
  @check muscl_omega in [-1.0, 0.0]

  """
  If `true`, a variable `u_err := u - ana_u` is defined and made available for 1d output.
  `u_err` is computed as the difference between the state variable `u` and an
  analytical reference solution `u_ana`, if available.
  """
  analyze_error = false
  @check analyze_error isa Bool

  """
  If `true`, then the variable `u_err_norm` is defined and made available for 0d output.
  `u_err_norm` is computed as the L^2-norm using the nodal approximation of the solution.
  Requires `analyze_error = true`.
  """
  analyze_error_norm = false
  @check analyze_error_norm isa Bool

  """
  If `true` any artificial viscosity related callbacks are run on every substep of the evolution.
  """
  av_recompute_substeps = false
  @check av_recompute_substeps isa Bool

end


end # ScalarEq
